#------------------------------------------------------------------------------
# ID of players/board side

# these must be 0,1 to allow direct array indexing (faster than using
# 'id == ID_PLAYER2' as the index)
ID_PLAYER1 = 0
ID_PLAYER2 = 1

ids = (ID_PLAYER1, ID_PLAYER2)

def IsIdValid(id):
	return id == ID_PLAYER1 or id== ID_PLAYER2

def IdOpponent(id):
	return 1-id		# see above about 0,1 assumption


#------------------------------------------------------------------------------
# parameters

# bin numbers are 1..BINS_PER_PLAYER and do not include the mancala.
BINS_PER_PLAYER = 6

def IsBinNumberValid(binNumber):
	return 1 <= binNumber and binNumber <= BINS_PER_PLAYER

BIN_NUMBER_MANCALA = 0


PIECES_PER_BIN = 4
TOTAL_PIECES = BINS_PER_PLAYER*2 * PIECES_PER_BIN

def IsAmountValid(amount):
	return amount >= 0


#------------------------------------------------------------------------------
# board

class Board:
	# no strict separation between bin numbers, mancala, and counter indices
	# because these routines are time-critical (called during eval).
	TOTAL_COUNTERS = (1+BINS_PER_PLAYER)*2
	idxMancalas = (0, BINS_PER_PLAYER+1)


	def __init__(self):
		self.counters = [PIECES_PER_BIN] * Board.TOTAL_COUNTERS
		for idSide in ids:
			self.counters[Board.idxMancalas[idSide]] = 0


	def AmountInMancala(self, idSide):
		idx = Board.idxMancalas[idSide]
		return self.counters[idx]

	def AddToMancala(self, idSide, difference):
		idx = Board.idxMancalas[idSide]
		self.counters[idx] += difference


	# also works for binNumber == BIN_NUMBER_MANCALA (simplifies
	# _DistributePieces)
	def AmountInBin(self, idSide, binNumber):
		idx = Board.idxMancalas[idSide]+binNumber
		return self.counters[idx]

	def AddToBin(self, idSide, binNumber, difference):
		#assert difference == 1 or difference == -1
		idx = Board.idxMancalas[idSide]+binNumber
		self.counters[idx] += difference

	def SetBinAmount(self, idSide, binNumber, amount):
		#assert IsAmountValid(amount)
		idx = Board.idxMancalas[idSide]+binNumber
		# setting the absolute counter value should only happen when
		# toggling between zero and another value.
		#assert amount == 0 or self.counters[idx] == 0
		self.counters[idx] = amount


	def VerifyTotalAmount(self):
		assert(sum(self.counters) == TOTAL_PIECES)

	def __eq__(self, other):
		for i in range(len(self.counters)):
			if self.counters[i] != other.counters[i]:
				return False
		return True

	def __ne__(self, other):
		return not self.__eq__(other)


#------------------------------------------------------------------------------
# base class for human/AI players

class Player:
	namesInUse = []

	def __init__(self, name):
		if name in Player.namesInUse:
			name += "2"
		self.name = name
		Player.namesInUse.append(name)

	def Name(self):
		return self.name

	def ChooseMove(self, idSide, board):
		assert 0	# abstract


#------------------------------------------------------------------------------
# game rules

def LegalMoves(idSide, board):
	""" return a list of all valid moves for <idSide>.
	"""
	return [binNumber for binNumber in range(1, BINS_PER_PLAYER+1) if board.AmountInBin(idSide, binNumber) != 0]


def _DistributePieces(idSide, binNumber, board):
	amount = board.AmountInBin(idSide, binNumber)
	if amount == 0:
		raise ValueError()
	board.SetBinAmount(idSide, binNumber, 0)

	idSide2 = idSide
	binNumber2 = binNumber-1
	while True:
		if binNumber2 != BIN_NUMBER_MANCALA or idSide2 == idSide:
			board.AddToBin(idSide2, binNumber2, 1)
			amount -= 1
			if amount == 0: break	# on last iteration, break before changing binNumber!

		if binNumber2 == BIN_NUMBER_MANCALA:
			idSide2 = IdOpponent(idSide2)
			binNumber2 = BINS_PER_PLAYER
		else:
			binNumber2 -= 1

	board.VerifyTotalAmount()
	return (idSide2, binNumber2)


def _CaptureOppositePieces(idSide, idSide2, binNumber2, board):
	if idSide != idSide2:
		return
	if binNumber2 == BIN_NUMBER_MANCALA:
		return
	# bin wasn't empty before this move
	if board.AmountInBin(idSide2, binNumber2) != 1:
		return

	idOpponent = IdOpponent(idSide)
	binNumberOpposite = BINS_PER_PLAYER+1 - binNumber2
	amount = board.AmountInBin(idOpponent, binNumberOpposite)
	if amount == 0:
		return
	
	board.AddToMancala(idSide,     amount+1)		# +1 for the capturing piece
	board.SetBinAmount(idSide,     binNumber2, 0)
	board.SetBinAmount(idOpponent, binNumberOpposite, 0)
	board.VerifyTotalAmount()


# note: it is difficult to undo captures; moves are taken back by reverting
# the entire game state. we therefore don't need to undo parts of a move.

def CarryOutMove(idSide, binNumber, board):
	""" return whose turn it is after this move """
	#assert IsIdValid(idSide)
	#assert IsBinNumberValid(binNumber)

	idSide2, binNumber2 = _DistributePieces(idSide, binNumber, board)

	_CaptureOppositePieces(idSide, idSide2, binNumber2, board)

	if idSide2 == idSide and binNumber2 == BIN_NUMBER_MANCALA:
		return idSide		# move again
	else:
		return IdOpponent(idSide)


def _AssignRemainingPiecesToMancalas(board):
	for idSide in ids:
		for binNumber in range(1, BINS_PER_PLAYER+1):
			amount = board.AmountInBin(idSide, binNumber)
			board.SetBinAmount(idSide, binNumber, 0)
			board.AddToMancala(idSide, amount)			


def AssignRemainingPiecesAndDetermineWinner(board):
	_AssignRemainingPiecesToMancalas(board)
	difference = board.AmountInMancala(ID_PLAYER1) - board.AmountInMancala(ID_PLAYER2)
	if difference > 0:
		return ID_PLAYER1
	elif difference < 0:
		return ID_PLAYER2
	else:
		return None
